<?php

namespace Luticate\Utils\Controller;

use FastRoute\Dispatcher;
use Luticate\Utils\Business\LuBusinessException;
use Luticate\Utils\Business\LuInjector;
use Luticate\Utils\Business\LuMethodDocParser;
use Luticate\Utils\Dbo\LuDboConstraintException;
use Luticate\Utils\Dbo\LuDboDeserializeException;
use FastRoute\DataGenerator\GroupCountBased as DataGeneratorGroupCountBased;
use FastRoute\Dispatcher\GroupCountBased as DispatcherGroupCountBased;
use FastRoute\RouteCollector;
use FastRoute\RouteParser\Std;
use Luticate\Utils\Dbo\LuMiddlewareDbo;
use Luticate\Utils\Dbo\LuRouteDbo;
use Luticate\Utils\Middleware\LuAbstractMiddleware;

class LuRoute {

    const REG_UINT = "[0-9]+";
    const REG_INT = "\\-?" . self::REG_UINT;
    const REG_BOOL = "true|false";

    /**
     * @var DispatcherGroupCountBased
     */
    private $_dispatcher = null;

    /**
     * @var LuMiddlewareDbo[]
     */
    private $_middlewares = [];

    /**
     * @var LuRouteDbo[]
     */
    private $routes = array();

    public function __construct()
    {
    }

    public function getRoutes()
    {
        return $this->routes;
    }

    /**
     * @param $middleware mixed
     */
    public function addMiddleware($middleware)
    {
        if (is_string($middleware)) {
            $this->_middlewares[] = new LuMiddlewareDbo($middleware);
        }
        else if (is_array($middleware)) {
            $this->_middlewares[] = LuMiddlewareDbo::jsonDeserialize($middleware);
        }
        else {
            $this->_middlewares[] = $middleware;
        }
    }
    
    public function setup()
    {
        $routeCollector = new RouteCollector(new Std(), new DataGeneratorGroupCountBased());
        $router = $this;

        foreach ($this->getRoutes() as $route) {
            $routeCollector->addRoute($route->getMethod(), $route->getUrl(), function($parameters) use($route, $router)
            {
                return $router->execute($route, $parameters);
            });
        }

        $this->_dispatcher = new DispatcherGroupCountBased($routeCollector->getData());
    }

    public static function getHeaders()
    {
        if (!function_exists('getallheaders')) {
            $out = [];
            foreach ($_SERVER as $key => $value) {
                if (substr($key, 0, 5) == "HTTP_") {
                    $key = str_replace(" ", "-", ucwords(strtolower(str_replace("_", " ", substr($key, 5)))));
                    $out[$key] = $value;
                }
                else {
                    $out[$key] = $value;
                }
            }
            return $out;
        }
        else {
            return getallheaders();
        }
    }

    public function execute(LuRouteDbo $route, $parameters)
    {
        $parameters["_headers"] = static::getHeaders();

        $middlewares = [];
        /**
         * @var $middlewareDbo LuMiddlewareDbo
         */
        foreach (array_merge($this->_middlewares, $route->getMiddlewares()) as $middlewareDbo) {
            /**
             * @var $middleware LuAbstractMiddleware
             */
            $middlewareName = $middlewareDbo->getMiddleware();
            $middleware = new $middlewareName();

            $args = array_merge($middlewareDbo->getParameters(), $route->getParameters());
            $args["_route"] = $route;
            $args["_headers"] = $parameters["_headers"];
            $args["_parameters"] = $parameters;
            $args["_all"] = $args;

            if (is_callable([$middleware, "onBefore"])) {
                $parameters = LuInjector::inject($middleware, "onBefore", $args);
            }

            $middlewares[] = [
                'middleware' => $middleware,
                'dbo' => $middlewareDbo
            ];
        }

        $result = LuInjector::inject($route->getController(), $route->getAction(), $parameters);

        foreach ($middlewares as $middlewareArray) {
            $middleware = $middlewareArray['middleware'];
            $middlewareDbo = $middlewareArray['dbo'];

            $args = array_merge($middlewareDbo->getParameters(), $route->getParameters());
            $args["_route"] = $route;
            $args["_headers"] = $parameters["_headers"];
            $args["_parameters"] = $parameters;
            $args["_result"] = $result;
            $args["_all"] = $args;

            if (is_callable([$middleware, "onAfter"])) {
                $result = LuInjector::inject($middleware, "onAfter", $args);
            }
        }

        return $result;
    }

    public function dispatch(string $httpMethod, string $url, $parameters)
    {
        $routeInfo = $this->_dispatcher->dispatch($httpMethod, $url);

        if ($routeInfo[0] == Dispatcher::NOT_FOUND) {
            throw new LuBusinessException("Route not found", 404);
        }
        else if ($routeInfo[0] == Dispatcher::METHOD_NOT_ALLOWED) {
            throw new LuBusinessException("Method not allowed", 405);
        }
        else {
            $handler = $routeInfo[1];
            $urlVars = $routeInfo[2];
            $parameters = array_merge($parameters, $urlVars);
            return $handler($parameters);
        }
    }
    
    public function addRoute2(LuRouteDbo $route) {
        $controller = $route->getController();
        if (strpos($controller, "\\") === false) {
            $controller = "App\\Controller\\" . $controller . "Controller";
        }
        $route->setController($controller);
        $this->routes[] = $route;
    }

    public function addRoute(string $httpMethod, string $url, string $controller, string $method, $parameters = array(), $middlewares = array())
    {
        $middlewaresDbo = [];
        foreach ($middlewares as $middleware) {
            if (is_string($middleware)) {
                $middlewaresDbo[] = new LuMiddlewareDbo($middleware);
            }
            else if (is_array($middleware)) {
                $middlewaresDbo[] = LuMiddlewareDbo::jsonDeserialize($middleware);
            }
            else {
                $middlewaresDbo[] = $middleware;
            }
        }
        
        $route = new LuRouteDbo();
        $route->setUrl($url);
        $route->setController($controller);
        $route->setAction($method);
        $route->setMiddlewares($middlewaresDbo);
        $route->setParameters($parameters);
        $route->setMethod($httpMethod);
        
        $this->addRoute2($route);
    }
    
    public function get(string $url, string $controller, string $method, $parameters = array(), $middleware = array())
    {
        $this->addRoute("GET", $url, $controller, $method, $parameters, $middleware);
    }

    public function post(string $url, string $controller, string $method, $parameters = array(), $middleware = array())
    {
        $this->addRoute("POST", $url, $controller, $method, $parameters, $middleware);
    }

    public function put(string $url, string $controller, string $method, $parameters = array(), $middleware = array())
    {
        $this->addRoute("PUT", $url, $controller, $method, $parameters, $middleware);
    }

    public function delete(string $url, string $controller, string $method, $parameters = array(), $middleware = array())
    {
        $this->addRoute("DELETE", $url, $controller, $method, $parameters, $middleware);
    }
}