<?php
/**
 * Created by PhpStorm.
 * User: robin
 * Date: 6/4/16
 * Time: 10:06 PM
 */

namespace Luticate\Utils\Controller;

use Luticate\Utils\Business\LuBusinessException;
use Luticate\Utils\Business\LuLog;
use Luticate\Utils\Dbo\LuRequestDbo;
use Ratchet\ConnectionInterface;
use Ratchet\MessageComponentInterface;
use Ratchet\Http\HttpServer;
use Ratchet\Server\IoServer;
use Ratchet\WebSocket\WsServer;

class LuticateApplication implements MessageComponentInterface
{
    private $_version = 1.0;
    private $_config = null;
    private $_clients = null;
    private $_router = null;
    
    private static $_app = null;
    
    public static function getInstance()
    {
        return self::$_app;
    }

    public static function getDatabase($name, $databases)
    {
        foreach ($databases as $database) {
            if (is_array($database) && $database["name"] == $name && !isset($database["alias"])) {
                return $database;
            }
        }
        return null;
    }

    public static function resolveDatabases($databases) {
        $dbs = [];
        foreach ($databases as $database) {
            if (isset($database["alias"])) {
                $db = static::getDatabase($database["alias"], $databases);
                unset($database["alias"]);
                if (!is_null($db)) {
                    $database = array_merge($db, $database);
                }
                else {
                    $database = null;
                }
            }
//            if (is_string($database)) {
//                $name = $database;
//                $database = static::getDatabase($name, $databases);
//                if (!is_null($database)) {
//                    $database["name"] = $name;
//                }
//            }
            if (is_array($database)) {
                $dbs[] = $database;
            }
        }
        return $dbs;
    }

    /**
     * @param $configuration array
     */
    public function __construct($configuration)
    {
        if (php_sapi_name() != "cli") {
            error_reporting(0);
        }
        $this->_config = $configuration;

        $this->_config["databases"] = static::resolveDatabases($this->_config["databases"]);

        if (!isset($this->_config['settings'])) {
            $this->_config['settings'] = [];
        }

        $this->_clients = new \SplObjectStorage;
        $this->_router = new LuRoute();
        LuLog::setLogFilePath(__DIR__ . "/../../../../../../". $this->_config["logs"]);
        self::$_app = $this;
    }

    public function getSetting($name, $default = null)
    {
        $value = getenv("lu_setting_" . $name);
        if ($value === false) {
            if (isset($this->_config['settings'][$name])) {
                $value = $this->_config['settings'][$name];
            }
            else {
                $value = $default;
            }
        }
        return $value;
    }

    /**
     * @return LuRoute
     */
    public function getRouter()
    {
        return $this->_router;
    }

    /**
     * @param LuRoute|null $router
     */
    public function setRouter($router)
    {
        $this->_router = $router;
    }

    /**
     * @param $version float
     */
    public function setVersion($version)
    {
        $this->_version = $version;
    }

    /**
     * @return float
     */
    public function getVersion()
    {
        return $this->_version;
    }
    
    public function setupRoutes()
    {
        /** @noinspection PhpUnusedLocalVariableInspection */
        $app = $this;
        /** @noinspection PhpIncludeInspection */
        require_once __DIR__ . '/../../../../../../app/routes.php';
        
        $this->_router->setup();
    }
    
    public function addDatabaseProvider($dbProviderClass)
    {
        call_user_func([$dbProviderClass, "setupDatabases"], $this->_config['databases']);
    }
    
    private function dispatch($httpMethod, $url, $parameters)
    {
        $r = new LuRequestDbo();
        $r->setVersion($this->getVersion());

        try {
            $result = $this->_router->dispatch($httpMethod, $url, $parameters);
            $r->setData($result);
            $r->setStatusCode(200);
        }
        catch (LuBusinessException $e)
        {
            LuLog::log($e);
            $r->setStatusCode($e->getCode());
            $r->setMessage($e->getMessage());
        }
        catch (\Exception $e)
        {
            LuLog::log($e);
            $r->setStatusCode(500);
            $r->setMessage("Internal Error");
        }
        catch (\Error $e)
        {
            LuLog::log($e);
            $r->setStatusCode(500);
            $r->setMessage("Internal Error");
        }
        return $r;
    }
    
    public function runHttp()
    {
        $httpMethod = $_SERVER['REQUEST_METHOD'];
        $url = strtok($_SERVER["REQUEST_URI"],'?');
        $parameters = array_merge($_GET, $_POST);
        
        $r = $this->dispatch($httpMethod, $url, $parameters);
        http_response_code($r->getStatusCode());
        header("content-type: application/json");
        echo $r->__toString();
    }
    
    public function runWs()
    {
        $server = IoServer::factory(
            new HttpServer(
                new WsServer(
                    $this
                )
            ),
            $this->_config['websocket']['port'],
            $this->_config['websocket']['address']
        );

        $server->run();
    }

    function onMessage(ConnectionInterface $from, $msg)
    {
        // TODO: Implement onMessage() method.
    }

    function onOpen(ConnectionInterface $conn)
    {
        $this->_clients->attach($conn);
    }

    function onClose(ConnectionInterface $conn)
    {
        $this->_clients->detach($conn);
    }

    function onError(ConnectionInterface $conn, \Exception $e)
    {
        $conn->close();
    }
}